'+--------------------------------------------------------------------------+
'|                               ObjReader                                  |
'|                                                                          |
'|             Load Model OBJ into an OpenGL/GDImage application.           |
'|                                                                          |
'+--------------------------------------------------------------------------+
'|                                                                          |
'|                         Author Patrice TERRIER                           |
'|                                                                          |
'|                           copyright (c) 2012                             |
'|                                                                          |
'|                Patrice Terrier http://www.zapsolution.com                |
'|                                                                          |
'+--------------------------------------------------------------------------+
'|                  Project started on : 06-30-2012 (MM-DD-YYYY)            |
'|                        Last revised : 07-30-2012 (MM-DD-YYYY)            |
'+--------------------------------------------------------------------------+
#COMPILE EXE "ObjReader.exe"

#INCLUDE "Win32API.inc"
#INCLUDE "GDImage.inc"
#INCLUDE "ZGLEXT.inc"
#INCLUDE "Mobj_dll.inc"
'#INCLUDE "Mobj.inc"
#LINK "zSpinPB10.sll"
#RESOURCE "objreader.pbr"

' WinXP Theme support declares (if applicable)
DECLARE FUNCTION EnableDialogTheme(BYVAL hDlg AS DWORD, BYVAL dwStyle AS DWORD) AS LONG
DECLARE FUNCTION IsThemeActive() AS LONG

DECLARE FUNCTION zTrace LIB "zTrace.DLL" ALIAS "zTrace" (zMessage AS ASCIIZ) AS LONG

MACRO CAMERA_FOVY  = 10
MACRO CAMERA_ZFAR  = 100
MACRO CAMERA_ZNEAR = 1

MACRO MOUSE_ORBIT_SPEED = 0.30     '// 0 = SLOWEST, 1 = FASTEST
MACRO MOUSE_DOLLY_SPEED = 0.02     '// same as above...but much more sensitive
MACRO MOUSE_TRACK_SPEED = 0.005    '// same as above...but much more sensitive

%CAMERA_NONE  = 0
%CAMERA_TRACK = 1
%CAMERA_DOLLY = 2
%CAMERA_ORBIT = 3

%ClientW       = 956
%ClientH       = 600

%ID_CTRL       = 101
%ID_LISTBOX    = 103
%ID_STATIC     = 112

%MENU_FILE_OPEN                 = 40004
%MENU_FILE_CLOSE                = 40005
%MENU_FILE_EXIT                 = 40006
%MENU_VIEW_FULLSCREEN           = 40012
%MENU_VIEW_RESET                = 40013
%MENU_VIEW_TEXTURED             = 40014
%MENU_VIEW_WIREFRAME            = 40015
%MENU_VIEW_CULLBACKFACES        = 40017
%MENU_SPECULAR_MODE             = 40019
%MENU_FIXED_PIPELINE            = 40021
%MENU_SWAP_VECTOR               = 40023

GLOBAL gn_GLwnd, gn_hMain AS DWORD, gr_Percent AS SINGLE
GLOBAL mt() AS ZGLTEXTUREX, texObj() AS ZGLTEXTUREX

GLOBAL gr_heading, gr_pitch, gr_MaxAnisotrophy AS SINGLE
GLOBAL gr_cameraPos(), gr_targetPos() AS SINGLE

GLOBAL gn_SupportsProgrammablePipeline, gn_NormalMappingShader, gn_BlinnPhongShader, gn_NullTexture, _
       gn_WindowWidth, gn_WindowHeight, gn_CullBackFaces, gn_EnableWireframe, gn_IsFullScreen,_
       gn_ObjectLoaded, gn_SpecularMode, gn_SwapVector AS LONG

FUNCTION zDefaultFont() AS LONG
    STATIC hDefault AS LONG
    IF hDefault = 0 THEN hDefault = GetStockObject(%ANSI_VAR_FONT)
    FUNCTION = hDefault
END FUNCTION

SUB zSetCTLFont(BYVAL hC AS LONG, BYVAL hFont AS LONG)
    CALL SendMessage(hC, %WM_SETFONT, hFont, 0)
END SUB

FUNCTION zStaticCenter(BYVAL Label$, BYVAL x AS LONG, BYVAL y AS LONG, BYVAL w AS LONG, BYVAL h AS LONG, BYVAL hParent AS LONG, BYVAL ID AS LONG) AS LONG
    LOCAl hCtl AS LONG
    hCtl = CreateWindowEx(0, "STATIC", (Label$), %WS_CHILD OR %WS_VISIBLE OR %WS_BORDER OR %SS_CENTER, _
                          x, y, w, h, hParent, ID, zInstance(), BYVAL %NULL)
    CALL zSetCTLFont(hCtl, zDefaultFont)
    FUNCTION = hCtl
END FUNCTION

'!'' We need this to share UseFont among procedures 
'!'SUB GlobalFont(UseFont AS ZGLFONT, BYVAL RW AS LONG)
'!'    STATIC WasFont AS ZGLFONT
'!'    IF RW THEN WasFont = UseFont
'!'    UseFont = WasFont
'!'END SUB

FUNCTION WINMAIN (BYVAL hInstance     AS LONG, _
                  BYVAL hPrevInstance AS LONG, _
                  BYVAL lpCmdLine     AS ASCIIZ PTR, _
                  BYVAL iCmdShow      AS LONG) AS LONG

    LOCAL Msg         AS tagMsg
    LOCAL wc          AS WndClassEx
    LOCAL zClass      AS ASCIIZ * 80
    LOCAL rc          AS RECT
    LOCAL x, y, IsInitialized AS LONG
    LOCAL dwStyle, dwExStyle, hMutex, hFound, hList AS DWORD
'
    LOCAL hWinXP_Lib           AS LONG ' Handle to WinXP Theme DLL
    LOCAL hWinXP_IsThemeActive AS LONG ' Handle to WinXP's IsThemeActive function   
'
'!'    DIM UseFont AS ZGLFONT 
'    
    zClass = "WavefrontOBJ"
'
    hMutex = CreateMutex(BYVAL %Null, 0, zClass)
    IF hMutex THEN
       IF GetLastError = %ERROR_ALREADY_EXISTS THEN
          DO
             hFound = FindWindow(zClass, ""): IF hFound THEN EXIT DO
             WHILE PeekMessage(Msg, %NULL, %NULL, %NULL, %PM_REMOVE): WEND
          LOOP
          IF IsIconic(hFound) THEN CALL ShowWindow(hFound, %SW_RESTORE)
          CALL SetForeGroundWindow(hFound)
          FUNCTION = 0
          EXIT FUNCTION
       END IF
    END IF
'
    wc.cbSize = SIZEOF(wc)
    IsInitialized = GetClassInfoEx(zInstance, zClass, wc)
    IF IsInitialized&   = 0 THEN
       wc.style         = %CS_HREDRAW OR %CS_VREDRAW
       wc.lpfnWndProc   = CODEPTR(WndProc)
       wc.cbClsExtra    = 0
       wc.cbWndExtra    = 0
       wc.hInstance     = zInstance()
       wc.hIcon         = LoadIcon(wc.hInstance, "PROGRAM")
       wc.hCursor       = LoadCursor(%NULL, BYVAL %IDC_ARROW)
       wc.hbrBackground = 16 ' %COLOR_BTNFACE + 1
       wc.lpszMenuName  = %NULL
       wc.lpszClassName = VARPTR(zClass)
       wc.hIconSm       = wc.hIcon
       IF RegisterClassEx(wc) THEN IsInitialized = %TRUE
    END IF
'
    IF IsInitialized THEN

     ' Load the WinXP Theme support (if applicable)
       hWinXP_Lib = LoadLibrary("UxTheme.dll")
       IF hWinXP_Lib THEN
          hWinXP_IsThemeActive = GetProcAddress(hWinXP_Lib, "IsThemeActive")
       END IF
'
     ' Window Extended Style
       dwExStyle = %WS_EX_APPWINDOW OR %WS_EX_WINDOWEDGE
     ' Windows Style
       dwStyle = %WS_OVERLAPPEDWINDOW
       CALL SetRect(rc, 0, 0, %ClientW, %ClientH)
       CALL AdjustWindowRectEx(rc, dwStyle, %FALSE, dwExStyle)  ' Adjust Window To True Requested Size
'
       x = MAX&((GetSystemMetrics(%SM_CXSCREEN) - rc.nRight - rc.nLeft) \ 2, 0)
       y = MAX&((GetSystemMetrics(%SM_CYSCREEN) - rc.nBottom - rc.nTop) \ 2, 0)
       MinTrackSizeW(rc.nRight - rc.nLeft)
       MinTrackSizeH(rc.nBottom - rc.nTop)
'
     ' Create The Window
       MyTitle$ = "GDImage control " + ZI_Version + " - ""OpenGL wavefront OBJ loader"" demo"
       gn_hMain = CreateWindowEx(dwExStyle, _        ' Extended Style For The Window
                             zClass, _               ' Class Name
                             (MyTitle$), _           ' Window Title
                             dwStyle OR _            ' Defined Window Style
                             %WS_CLIPSIBLINGS OR _   ' Required Window Style
                             %WS_CLIPCHILDREN, _     ' Required Window Style
                             x, y, _                 ' Window Position
                             MinTrackSizeW(0), _     ' Calculate Window Width
                             MinTrackSizeH(0), _     ' Calculate Window Height
                             %NULL, _                ' No Parent Window
                             %NULL, _                ' No Menu
                             wc.hInstance, _         ' Instance
                             BYVAL %NULL)            ' Dont Pass Anything To WM_CREATE
'
       IF gn_hMain THEN

'         // Apply WinXP Theme support
          LOCAL lRes AS LONG, pProc AS DWORD
          IF hWinXP_IsThemeActive THEN
             CALL DWORD hWinXP_IsThemeActive USING IsThemeActive TO lRes
             IF lRes  THEN pProc = GetProcAddress(lRes, "EnableThemeDialogTexture")
             IF pProc THEN CALL DWORD pProc USING EnableDialogTheme(gn_hMain, &H01 OR &H02 OR &H04 OR &H06)
          END IF

'         // Create Static.
          CALL zStaticCenter($CR + "Left mouse button:" + $CR + "change view angle." + $CR + _
                             $CR + "Right mouse button:" + $CR + "change graph location." + $CR + _
                             $CR + "Mouse wheel zooming.", _
                             %ClientW - (8 + 120 + 8), 19, 120, 120, gn_hMain, %ID_STATIC)
          CALL ZI_SetAnchorMode(GetDlgItem(gn_hMain, %ID_STATIC), %ANCHOR_RIGHT)

'         // Create ListBox to select texture from.
          hList = CreateWindowEx(0, "ListBox", "", _
                             %WS_CHILD OR %WS_BORDER OR %WS_VISIBLE OR %WS_VSCROLL OR %LBS_SORT OR _
                             %LBS_NOTIFY OR %WS_TABSTOP OR %LBS_HASSTRINGS OR %LBS_DISABLENOSCROLL , _
                             %ClientW - (8 + 120 + 8), 19 + (22 + 5) * 5 + 5, 120, 120, _
                             gn_hMain, %ID_LISTBOX, zInstance(), BYVAL %NULL)
          CALL ZI_SetAnchorMode(hList, %ANCHOR_RIGHT)
          CALL zSetCTLFont(hList, zDefaultFont())
          gl_LoadBackground(hList)
 
'         *******************************************************************************
'         // 6.04 Create wavefront OpenGL control.
'         -------------------------------------------------------------------------------
          dwStyle = %WS_CHILD OR %WS_VISIBLE
          dwExStyle = 0
          gn_GLwnd = Mobj_CreateWindow(dwExStyle, dwStyle, 0, 0, 800, %ClientH, gn_hMain, %ID_CTRL)

          IF gn_GLwnd THEN
             ZI_SetAnchorMode(gn_GLwnd, %ANCHOR_HEIGHT_WIDTH)  ' Anchor the control (make it a resizable)

             InitGL()
             InitApp()

'!''            // Create a WGL font for the XYZ axis.
'!'             ZI_BuildGLfont(ZI_GetGLDC(gn_GLwnd), UseFont) ' Build OpenGL font for our OpenGL window
'!'             globalFont(UseFont, 1)

             DIM mt(1 TO 2) AS ZGLTEXTUREX
             mt(1).FullName = EXE.Path$ + ListGetText(hList, ListGetCursel(hList)): mt(1).ID = 1: mt(1).Square = 0
             'mt(2).FullName = EXE.Path$ + "texture.jpg": mt(2).ID = 2: mt(2).Square = 0
             ZI_MakeMultipleTexture(BYVAL VARPTR(mt(LBOUND(mt))), UBOUND(mt) - LBOUND(mt) + 1)
            
             LOCAL hMnu, hSubMenu, hSubMenu2 AS DWORD
             hMnu = CreateMenu()
                hSubMenu = CreatePopUpMenu()
                AppendMenu (hMnu, %MF_POPUP OR %MF_ENABLED, hSubMenu, "&File")
                   AppendMenu (hSubMenu, %MF_ENABLED, %MENU_FILE_OPEN, "&Open...")
                   AppendMenu (hSubMenu, %MF_ENABLED, %MENU_FILE_CLOSE, "&Close")
                   AppendMenu (hSubMenu, %MF_SEPARATOR, 0, "")
                   AppendMenu (hSubMenu, %MF_ENABLED, %MENU_FILE_EXIT, "E&xit")
                hSubMenu = CreatePopUpMenu()
                AppendMenu (hMnu, %MF_POPUP OR %MF_ENABLED, hSubMenu, "&View")
                   AppendMenu (hSubMenu, %MF_ENABLED, %MENU_VIEW_FULLSCREEN, "&Fullscreen")
                   AppendMenu (hSubMenu, %MF_ENABLED, %MENU_VIEW_RESET, "&Reset")
                   AppendMenu (hSubMenu, %MF_SEPARATOR, 0, "")
                   'AppendMenu (hSubMenu, %MF_CHECKED, %MENU_VIEW_CULLBACKFACES, "&Cull back faces")
                   AppendMenu (hSubMenu, %MF_ENABLED, %MENU_VIEW_CULLBACKFACES, "&Cull back faces")
                   'AppendMenu (hSubMenu, %MF_CHECKED, %MENU_VIEW_TEXTURED, "&Textured")
                   AppendMenu (hSubMenu, %MF_ENABLED, %MENU_VIEW_TEXTURED, "&Textured")
                   AppendMenu (hSubMenu, %MF_ENABLED, %MENU_VIEW_WIREFRAME, "&Wireframe")
                   AppendMenu (hSubMenu, %MF_SEPARATOR, 0, "")
                   IF gn_SupportsProgrammablePipeline THEN
                      'AppendMenu (hSubMenu, %MF_ENABLED, %MENU_FIXED_PIPELINE, "&Use fixed pipeline")
                      AppendMenu (hSubMenu, %MF_CHECKED, %MENU_FIXED_PIPELINE, "&Use fixed pipeline")
                   END IF
                   'AppendMenu (hSubMenu, %MF_ENABLED, %MENU_SPECULAR_MODE, "Use &Specular mode")
                   AppendMenu (hSubMenu, %MF_CHECKED, %MENU_SPECULAR_MODE, "Use &Specular mode")
                   AppendMenu (hSubMenu, %MF_SEPARATOR, 0, "")
                   AppendMenu (hSubMenu, %MF_ENABLED, %MENU_SWAP_VECTOR, "Swap &Vector direction")
             SetMenu (gn_hMain, hMnu)

'            // Show the main window
             CALL ShowWindow(gn_hMain, iCmdShow)
             CALL SetForegroundWindow(gn_hMain)                ' Slightly Higher Priority
             CALL SetFocus(gn_hMain)                           ' Sets Keyboard Focus To The Window

      
'            // Main message pump.
             CALL SetTimer(gn_hMain, 1, 0, 0) ' 66 FPS
             WHILE GetMessage(Msg, 0, 0, 0)
                   TranslateMessage(Msg)
                   DispatchMessage(Msg)
             WEND
             CALL KillTimer(gn_hMain, 1)

'            // CLEANUP
             Cleanup()
'!''            // Delete GLFont.
'!'             ZI_DeleteGLFont(UseFont)

             FUNCTION = msg.wParam
          END IF
       END IF

'      // UNLOAD the WinXP Theme DLL (if necessary)
       IF hWinXP_Lib THEN CALL FreeLibrary(hWinXP_Lib)

    END IF
'
    IF hMutex THEN CALL CloseHandle(hMutex)
'
END FUNCTION

SUB gl_DestroyTexture()
    LOCAL K, mtCount, OkDelete AS LONG
    mtCount = UBOUND(mt()) - LBOUND(mt()) + 1
    IF mtCount THEN
       DIM Texture(1 TO mtCount) AS LONG
       FOR K = 1 TO mtCount
           Texture(K) = mt(k).Texture: IF Texture(K) THEN OkDelete = -1
       NEXT
       IF OkDelete THEN CALL glDeleteTextures(mtCount, Texture(1))
    END IF
END SUB

SUB Cleanup()

    LOCAL K, mtCount, OkDelete AS LONG
    mtCount = UBOUND(mt()) - LBOUND(mt()) + 1
    IF mtCount THEN
       DIM Texture(1 TO mtCount) AS LONG
       FOR K = 1 TO mtCount
           Texture(K) = mt(k).Texture: IF Texture(K) THEN OkDelete = -1
       NEXT
       IF OkDelete THEN CALL glDeleteTextures(mtCount, Texture(1))
    END IF

'   //////////////
    Mobj_destroy()

    IF gn_NullTexture THEN
        glDeleteTextures(1, gn_NullTexture)
        gn_NullTexture = 0
    END IF

'   // Cleanup texture from memory             
    DestroyObjTexture()

    IF ZI_GetGLDC(gn_GLwnd) THEN
        IF ZI_GetGLRC(gn_GLwnd) THEN
            wglMakeCurrent(ZI_GetGLDC(gn_GLwnd), 0)
            wglDeleteContext(ZI_GetGLRC(gn_GLwnd))
            ZI_SetProperty(gn_GLwnd, %ZI_GLRC, 0)
        END IF
        ReleaseDC(gn_GLwnd, ZI_GetGLDC(gn_GLwnd))
        ZI_SetProperty(gn_GLwnd, %ZI_GLDC, 0)
    END IF
END SUB

SUB ResetCamera()

    REDIM gr_targetPos(2), gr_cameraPos(2) AS SINGLE

    Mobj_getCenter(gr_targetPos(0), gr_targetPos(1), gr_targetPos(2))

    gr_cameraPos(0) = gr_targetPos(0)
    gr_cameraPos(1) = gr_targetPos(1)
    gr_cameraPos(2) = 6 ' gr_targetPos(2) + Mobj_getRadius() + CAMERA_ZNEAR + 0.4

    gr_pitch = 0.0
    gr_heading = 0.0

END SUB

SUB InitGL()

    LOCAL glDC, glRC AS LONG
    glDC = GetDC(gn_GLwnd)
    IF glDC = 0 THEN Mobj_ReportError("GetDC() failed."): EXIT SUB

    LOCAL pf, msaaSamples AS LONG
    LOCAL pfd AS PIXELFORMATDESCRIPTOR
    LOCAL osvi AS OSVERSIONINFO

    pfd.nSize              = SIZEOF(pfd)
    pfd.nVersion           = 1
    pfd.dwFlags            = %PFD_DRAW_TO_WINDOW OR %PFD_SUPPORT_OPENGL OR %PFD_DOUBLEBUFFER
    pfd.iPixelType         = 0 ' PFD_TYPE_RGBA
    pfd.cColorBits         = 32
    pfd.cAlphaBits         = 1
    pfd.cDepthBits         = 16
    pfd.iLayerType         = 0 ' %PFD_MAIN_PLANE

    osvi.dwOSVersionInfoSize = SIZEOF(OSVERSIONINFO)
    IF GetVersionEx(osvi) THEN ' Check for VISTA and above
       IF osvi.dwPlatformId > 1 AND osvi.dwMajorVersion > 5 THEN pfd.dwFlags = pfd.dwFlags OR %PFD_SUPPORT_COMPOSITION
    END IF

    Mobj_ChoosePixelFormat(pf, msaaSamples)

    IF pf = 0 THEN pf = ChoosePixelFormat(glDC, pfd)

    IF SetPixelFormat(glDC, pf, pfd) THEN
       glRC = wglCreateContext(glDC): IF glRC = 0 THEN Mobj_ReportError("Unable to create wglCreateContext."): EXIT SUB
       IF wglMakeCurrent(glDC, glRC) = 0 THEN Mobj_ReportError("Unable to make wglMakeCurrent."): EXIT SUB
       ZI_SetProperty(gn_GLwnd, %ZI_GLDC, glDC)
       ZI_SetProperty(gn_GLwnd, %ZI_GLRC, glRC)

       gn_SupportsProgrammablePipeline = Mobj_GL2SupportsGLVersion(2, 0)

'      // Check for GL_EXT_texture_filter_anisotropic support.
       IF Mobj_WGLisExtensionSupported("GL_EXT_texture_filter_anisotropic") THEN
          glGetFloatv(%GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, gr_MaxAnisotrophy)
       ELSE
          gr_MaxAnisotrophy = 1.0
       END IF

    END IF

END SUB

SUB InitApp()

    gn_CullBackFaces = 0  '-1
    gn_SpecularMode = -1 'Mobj_enableTextures(-1, 1)
    gn_ObjectLoaded = 0 


    glClearColor(0.3, 0.5, 0.9, 0.0)
    glClearDepth(1)

    glEnable(%GL_TEXTURE_2D)
    glEnable(%GL_DEPTH_TEST)
    IF gn_CullBackFaces THEN glEnable(%GL_CULL_FACE)
    glEnable(%GL_LIGHTING)
    glEnable(%GL_LIGHT0)

    glActiveTexture(%GL_TEXTURE1)
    glEnable(%GL_TEXTURE_2D)
    glTexParameteri(%GL_TEXTURE_2D, %GL_TEXTURE_MAG_FILTER, %GL_LINEAR)
    glTexParameteri(%GL_TEXTURE_2D, %GL_TEXTURE_MIN_FILTER, %GL_LINEAR_MIPMAP_LINEAR)
    glTexParameteri(%GL_TEXTURE_2D, %GL_TEXTURE_WRAP_S, %GL_REPEAT)
    glTexParameteri(%GL_TEXTURE_2D, %GL_TEXTURE_WRAP_T, %GL_REPEAT)

    glActiveTexture(%GL_TEXTURE0)
    glEnable(%GL_TEXTURE_2D)
    glTexParameteri(%GL_TEXTURE_2D, %GL_TEXTURE_MAG_FILTER, %GL_LINEAR)
    glTexParameteri(%GL_TEXTURE_2D, %GL_TEXTURE_MIN_FILTER, %GL_LINEAR_MIPMAP_LINEAR)
    glTexParameteri(%GL_TEXTURE_2D, %GL_TEXTURE_WRAP_S, %GL_REPEAT)
    glTexParameteri(%GL_TEXTURE_2D, %GL_TEXTURE_WRAP_T, %GL_REPEAT)

    IF gn_SupportsProgrammablePipeline THEN
       gn_BlinnPhongShader = Mobj_LoadShader(%SHADER_BLINN_PHONG)
       IF gn_BlinnPhongShader = 0 THEN Mobj_ReportError("Failed to load Blinn-Phong shader.")

       gn_NormalMappingShader = Mobj_LoadShader(%SHADER_NORMAL_MAPPING)
       IF gn_NormalMappingShader = 0 THEN Mobj_ReportError("Failed to load normal mapping shader.")

       gn_NullTexture = Mobj_CreateNullTexture(2, 2)
       IF gn_NullTexture = 0 THEN Mobj_ReportError("Failed to create null texture.")
    END IF

END SUB

SUB DrawFrame()

    STATIC nFlip AS LONG
    nFlip = NOT nFlip: IF nFlip THEN EXIT SUB

    glClear(%GL_COLOR_BUFFER_BIT OR %GL_DEPTH_BUFFER_BIT)

    glLoadIdentity()

    LOCAL WasLighting AS LONG
    WasLighting = glIsEnabled(%GL_LIGHTING)
    IF WasLighting THEN CALL glDisable(%GL_LIGHTING)
    glEnable(%GL_TEXTURE_2D)
    glColor4ub(255,255,255,255)
    GL_BindTexture(%GL_TEXTURE_2D, mt(1).texture)
    GL_DrawQuad(CAMERA_FOVY * 1.867, CAMERA_FOVY * 0.9335, -99.99)
    IF WasLighting THEN CALL glEnable(%GL_LIGHTING)

    IF gn_ObjectLoaded THEN

        LOCAL rtX, rtY, rUpx, rUpy, rUpz AS SINGLE
        IF gn_SwapVector THEN
           rUpx = 1
           rtX = gr_targetPos(1)
           rtY = -gr_targetPos(0)
        ELSE
           rUpy = 1
           rtX = gr_targetPos(0)
           rtY = gr_targetPos(1)
        END IF
        gluLookAt(gr_cameraPos(0), gr_cameraPos(1), gr_cameraPos(2), rtX, rtY, gr_targetPos(2), _
                  rUpx, rUpy, rUpz)
     
        IF gn_SwapVector THEN
           glRotatef(gr_pitch, 0.0, 1.0, 0.0)
           glRotatef(gr_heading, 1.0, 0.0, 0.0)
        ELSE
           glRotatef(gr_pitch, 1.0, 0.0, 0.0)
           glRotatef(gr_heading, 0.0, 1.0, 0.0)
        END IF

        LOCAL UseFixedPipeline AS LONG
        UseFixedPipeline = GetMenuState(GetMenu(gn_hMain), %MENU_FIXED_PIPELINE, %MF_BYCOMMAND) AND %MF_CHECKED = %MF_CHECKED

        IF gn_SupportsProgrammablePipeline AND UseFixedPipeline = 0 THEN
           Mobj_DrawUsingProgrammablePipeline()
        ELSE

           IF gn_SpecularMode THEN GL_TexSpec(%TRUE)
              glEnable(%GL_BLEND)
              glBlendFunc(%GL_SRC_ALPHA, %GL_ONE_MINUS_SRC_ALPHA)

              GL_BindTexture(%GL_TEXTURE_2D, gn_NullTexture) 
'              GL_BindTexture(%GL_TEXTURE_2D, mt(2).texture)
              DrawUsingFixedFuncPipeline()
           IF gn_SpecularMode THEN GL_TexSpec(%FALSE)

        END IF

    END IF

'!''   // Draw zoom percentage
'!''   // Retrieve current WGL font
'!'    LOCAL UseFont AS ZGLFONT, zTxt AS ASCIIZ * 16
'!'    globalFont(UseFont, 0)
'!'    ZI_DrawGLText(gn_GLwnd, UseFont, 10, 10 + UseFont.fontHeight * 0, ("Zoom " + STR$(CLNG(gr_cameraPos(2) * 100))) + "%", ZD_ARGB(255, 255,255,0))

    glFlush()
    ZI_UpdateGLWindow(gn_GLwnd)

END SUB

FUNCTION WndProc(BYVAL hWnd AS LONG, BYVAL Msg AS LONG, BYVAL wParam AS LONG, BYVAL lParam AS LONG) AS LONG
    LOCAL hCtrl, wP AS LONG

    SELECT CASE LONG Msg

    CASE %WM_GETMINMAXINFO
         LOCAL pMM AS MINMAXINFO PTR
         pMM = lParam
         @pMM.ptMinTrackSize.x = MinTrackSizeW(0) '// minimum width size of the window.
         @pMM.ptMinTrackSize.y = MinTrackSizeH(0) '// minimum height size of the window.

    CASE %WM_TIMER
         DrawFrame()
         FUNCTION = 0: EXIT FUNCTION

    CASE %WM_MOUSEWHEEL
         IF GetFocus() <> gn_GLwnd THEN SetFocus(gn_GLwnd)

    CASE %WM_SIZE
         'ZI_ResizeGLWindow(gn_GLwnd)
         DrawFrame()

    CASE %WM_COMMAND
         wP = LOWRD(wParam)

         SELECT CASE LONG wP
  
         CASE %ID_LISTBOX
              IF HI(WORD, wParam) = %LBN_SELCHANGE THEN
                 hCtrl = GetDlgItem(hWnd, %ID_LISTBOX)
                 mt(1).FullName = EXE.Path$ + ListGetText(hCtrl, ListGetCursel(hCtrl))
                 ZI_UpdateNamedGLTextureFromFileEx(mt(1).FullName, mt(1).Texture, mt(1).Square)
                 gr_Percent = 0
              END IF
         CASE ELSE
              ProcessMenu(hWnd, wParam, lParam)
         END SELECT
         FUNCTION = 0: EXIT FUNCTION

    CASE %WM_DESTROY
         CALL PostQuitMessage(0)
         FUNCTION = 0: EXIT FUNCTION
    END SELECT
    FUNCTION = DefWindowProc(hWnd, Msg, wParam, lParam)

END FUNCTION

SUB ListDeleteAll(BYVAL hList AS LONG)
    IF hList THEN CALL SendMessage(hList, %LB_RESETCONTENT, 0, 0)
END SUB

SUB ListSelectPlus(BYVAL hList AS LONG, BYVAL nSelected AS LONG)
    IF nSelected > 0 THEN DECR nSelected '// Because ListBox is zero based
    IF hList THEN CALL SendMessage(hList, %LB_SETCURSEL, nSelected, 0)
END SUB

FUNCTION ListAdd(BYVAL hList AS LONG, zTmp AS ASCIIZ) AS LONG
    IF hList THEN
       IF LEN(zTmp) THEN FUNCTION = SendMessage(hList, %LB_ADDSTRING, 0, VARPTR(zTmp))
    END IF
END FUNCTION

FUNCTION ListGetCursel(BYVAL hList AS LONG) AS LONG
    LOCAL nSelected AS LONG
    IF hList THEN
       nSelected = SendMessage(hList, %LB_GETCURSEL, 0, 0)
       IF nSelected > -1 THEN INCR nSelected '// Because ListBox is zero based
    END IF
    FUNCTION = nSelected
END FUNCTION

FUNCTION ListGetText(BYVAL hList AS LONG, BYVAL Item AS LONG) AS STRING
    LOCAL sItem AS STRING, nLEN AS LONG
    IF Item > 0 THEN
       DECR Item
       nLEN = SendMessage(hList, %LB_GETTEXTLEN, Item, 0)
       sItem = SPACE$(nLEN)
       CALL SendMessage(hList, %LB_GETTEXT, Item, BYVAL STRPTR(sItem))
       FUNCTION = sItem
    END IF
END FUNCTION

SUB gl_LoadBackground(BYVAL hList AS LONG)
    IF hList = 0 THEN EXIT SUB

    LOCAL nCount, WashFind, hFind, nItem AS LONG
    LOCAL fd AS WIN32_FIND_DATA
    LOCAL zTmp AS ASCIIZ * %MAX_PATH
    LOCAL zExt AS ASCIIZ * 8

    ListDeleteAll(hList)

    WashFind = FindFirstFile(EXE.Path$ + "Back*.jpg", fd)
    hFind = WashFind
    nCount = 0
    DO WHILE hFind > 0
       IF ASC(fd.cFileName) <> 46 THEN
          IF INSTR(CHR$(0,1,32,33,128), CHR$(fd.dwFileAttributes)) THEN
             zTmp = fd.cFileName
             zExt = PATHNAME$(EXTN, fd.cFileName) + ".": CALL CharLower(zExt)
             IF INSTR(".jpg.png.bmp.", zExt) THEN
                IF ListAdd(hList, zTmp) > -1 THEN
                   nCount += 1
                   IF zTmp = "Back02.jpg" THEN nItem = nCount
                END IF
             END IF
          END IF
       END IF
       hFind = FindNextFile(WashFind, fd)
    LOOP
    CALL FindClose(WashFind)
    IF nCount > 0 THEN
       IF nItem = 0 THEN nItem = 1
       ListSelectPlus(hList, nItem)
    END IF
END SUB

'// Main window minimum horizontal track size
FUNCTION MinTrackSizeW ALIAS "MinTrackSizeW"(BYVAL SizeX AS LONG) AS LONG
    STATIC WasSizeX AS LONG
    IF SizeX THEN WasSizeX = SizeX
    FUNCTION = WasSizeX
END FUNCTION

'// Main window minimum vertical track size
FUNCTION MinTrackSizeH ALIAS "MinTrackSizeH"(BYVAL SizeY AS LONG) AS LONG
    STATIC WasSizeY AS LONG
    IF SizeY THEN WasSizeY = SizeY
    FUNCTION = WasSizeY
END FUNCTION

SUB DrawUsingFixedFuncPipeline()

    LOCAL pMesh AS Mesh PTR
    LOCAL pMaterial AS Material PTR
    LOCAL pVertices AS VertexT PTR
    LOCAL nI, nCount, nVertexSize AS LONG
    LOCAL nPtr AS DWORD

    nVertexSize = Mobj_getVertexSize()
    FOR nI = 0 TO Mobj_getNumberOfMeshes() - 1

        pMesh = Mobj_getMesh(nI)
        pMaterial = @pMesh.pMaterial
        pVertices = Mobj_getVertexBuffer()

        glMaterialfv(%GL_FRONT_AND_BACK, %GL_AMBIENT, @pMaterial.ambient(0))
        glMaterialfv(%GL_FRONT_AND_BACK, %GL_DIFFUSE, @pMaterial.diffuse(0))
        glMaterialfv(%GL_FRONT_AND_BACK, %GL_SPECULAR, @pMaterial.specular(0))
        glMaterialf(%GL_FRONT_AND_BACK, %GL_SHININESS, @pMaterial.shininess * 128.0)

        glEnable(%GL_TEXTURE_2D)
        IF Mobj_enableTextures(0, 0) THEN
           nCount = UBOUND(texObj())
           IF nCount THEN
              FOR nCount = LBOUND(texObj()) TO UBOUND(texObj())
                  IF texObj(nCount).FullName = @pMaterial.szColorMapFilename THEN
                     glBindTexture(%GL_TEXTURE_2D, texObj(nCount).texture) '  texObj(nCount).ID)
                     EXIT FOR
                  END IF
              NEXT
           ELSE
              GL_BindTexture(%GL_TEXTURE_2D, gn_NullTexture)
              'GL_BindTexture(%GL_TEXTURE_2D, mt(2).texture)
           END IF
        ELSE
           IF gn_SpecularMode THEN
              GL_BindTexture(%GL_TEXTURE_2D, mt(1).texture)
           ELSE
              glDisable(%GL_TEXTURE_2D)
           END IF
        END IF

        IF Mobj_hasPositions() THEN
            glEnableClientState(%GL_VERTEX_ARRAY)
            My_glVertexPointer(3, %GL_FLOAT, nVertexSize, Mobj_getVertexBufferPosition())
        END IF

        IF Mobj_hasTextureCoords() THEN
            glEnableClientState(%GL_TEXTURE_COORD_ARRAY)
            My_glTexCoordPointer(2, %GL_FLOAT, nVertexSize, Mobj_getVertexBufferTexCoord())
        END IF

        IF Mobj_hasNormals() THEN
            glEnableClientState(%GL_NORMAL_ARRAY)
            My_glNormalPointer(%GL_FLOAT, nVertexSize, Mobj_getVertexBufferNormal())
        END IF

        My_glDrawElements(%GL_TRIANGLES, @pMesh.triangleCount * 3, %GL_UNSIGNED_INT, Mobj_getIndexBuffer() + (@pMesh.startIndex * Mobj_getIndexSize()) )

        IF Mobj_hasNormals() THEN glDisableClientState(%GL_NORMAL_ARRAY)

        IF Mobj_hasTextureCoords() THEN glDisableClientState(%GL_TEXTURE_COORD_ARRAY)

        IF Mobj_hasPositions() THEN glDisableClientState(%GL_VERTEX_ARRAY)
    NEXT
END SUB

SUB LoadModel(pszFilename AS ASCIIZ)

    '// Import the OBJ file and normalize to unit length.

    'SetCursor(LoadCursor(%NULL, BYVAL %IDC_WAIT))
    zSpinnerInit(gn_GLwnd, EXE.Path$ + "ObjReader.ski", 0)

    'IF Mobj_import(pszFilename, 0) THEN
    'LAC Changed from 0 to 1: ' Some models didn't have good normals
    IF Mobj_import(pszFilename, 1) THEN
       Mobj_normalize(1.0, %TRUE)
    END IF

'''''''''''''''''''''''''''''''''''''''
    '// Load any associated textures.
    LOCAL pMaterial AS Material PTR
    LOCAL nI, nItem, nCount AS LONG
    REDIM sTexture(Mobj_getNumberOfMaterials()) AS STRING
    FOR nI = 0 TO Mobj_getNumberOfMaterials() - 1

        pMaterial = Mobj_getMaterial(nI)
        ARRAY SCAN sTexture(), = @pMaterial.szColorMapFilename, TO nItem
        IF nItem = 0 THEN 
           INCR nCount
           sTexture(nCount) = @pMaterial.szColorMapFilename
           nItem = nCount
        END IF

    NEXT
    DestroyObjTexture()   
    IF nCount THEN
       REDIM texObj(1 TO nCount) AS ZGLTEXTUREX
       FOR nI = 1 TO nCount
           texObj(nI).FullName = sTexture(nI): texObj(nI).ID = nI + 100: texObj(nI).Square = 0
       NEXT
       ZI_MakeMultipleTexture(BYVAL VARPTR(texObj(LBOUND(texObj))), UBOUND(texObj) - LBOUND(texObj) + 1)
       Mobj_CopyMultipleTexture(BYVAL VARPTR(texObj(LBOUND(texObj))), UBOUND(texObj) - LBOUND(texObj) + 1)
       ERASE sTexture()
    END IF
'''''''''''''''''''''''''''''''''''''''

    SetWindowText(gn_hMain, MCASE$(PATHNAME$(NAMEX, pszFileName)))
    gn_ObjectLoaded = -1

    'SetCursor(LoadCursor(%NULL, BYVAL %IDC_ARROW))
    IF zSpinnerRunning() THEN zSpinnerClose()

END SUB

SUB DestroyObjTexture()
    LOCAL K, texObjCount, OkDelete AS LONG
    texObjCount = UBOUND(texObj()) - LBOUND(texObj()) + 1
    IF texObjCount THEN
       REDIM Texture(1 TO texObjCount) AS LONG
       FOR K = 1 TO texObjCount
           Texture(K) = texObj(k).Texture: IF Texture(K) THEN OkDelete = -1
       NEXT
       IF OkDelete THEN CALL glDeleteTextures(texObjCount, Texture(1))
    END IF
    ERASE texObj()
END SUB

SUB ToggleFullScreen()

    STATIC savedExStyle, savedStyle, windowWidth, windowHeight AS DWORD
    STATIC rcSaved AS RECT

    gn_IsFullScreen = NOT gn_IsFullScreen

    IF gn_IsFullScreen THEN
'       // Moving to full screen mode.
        savedExStyle = GetWindowLong(gn_hMain, %GWL_EXSTYLE)
        savedStyle = GetWindowLong(gn_hMain, %GWL_STYLE)
        GetWindowRect(gn_hMain, rcSaved)

        SetWindowLong(gn_hMain, %GWL_EXSTYLE, 0)
        SetWindowLong(gn_hMain, %GWL_STYLE, %WS_POPUP OR %WS_CLIPCHILDREN OR %WS_CLIPSIBLINGS)
        SetWindowPos(gn_hMain, %HWND_TOPMOST, 0, 0, 0, 0, %SWP_NOMOVE OR %SWP_NOSIZE OR %SWP_FRAMECHANGED OR %SWP_SHOWWINDOW)

        windowWidth = GetSystemMetrics(%SM_CXSCREEN)
        windowHeight = GetSystemMetrics(%SM_CYSCREEN)

        SetWindowPos(gn_hMain, %HWND_TOPMOST, 0, 0, windowWidth, windowHeight, %SWP_SHOWWINDOW)
    
    ELSE
'       // Moving back to windowed mode.

        SetWindowLong(gn_hMain, %GWL_EXSTYLE, savedExStyle)
        SetWindowLong(gn_hMain, %GWL_STYLE, savedStyle)
        SetWindowPos(gn_hMain, %HWND_NOTOPMOST, 0, 0, 0, 0, %SWP_NOMOVE OR %SWP_NOSIZE OR %SWP_FRAMECHANGED OR %SWP_SHOWWINDOW)

        windowWidth = rcSaved.right - rcSaved.left
        windowHeight = rcSaved.bottom - rcSaved.top

        SetWindowPos(gn_hMain, %HWND_NOTOPMOST, rcSaved.left, rcSaved.top, windowWidth, windowHeight, %SWP_SHOWWINDOW)
    END IF

END SUB

SUB ProcessMenu(BYVAL hWnd AS DWORD , BYVAL wParam AS LONG, BYVAL lParam AS LONG)

    STATIC szFile AS ASCIIZ * %MAX_PATH
    STATIC ofn AS OPENFILENAME

    SELECT CASE LONG LO(WORD, wParam)

    CASE %MENU_FILE_OPEN:
        ofn.lStructSize = SIZEOF(ofn)
        ofn.hwndOwner = hWnd
        LOCAL sFilter, sTitle AS STRING
        sFilter = "Alias|Wavefront (*.OBJ)" + $NUL + "*.obj" + $NUL
        ofn.lpstrFilter = STRPTR(sFilter)
        ofn.lpstrCustomFilter = 0
        ofn.nFilterIndex = 1
        ofn.lpstrFile = VARPTR(szFile)
        ofn.nMaxFile = %MAX_PATH
        sTitle = "Open File"
        ofn.lpstrTitle = STRPTR(sTitle)
        ofn.lpstrFileTitle = 0
        ofn.lpstrDefExt = 0
        ofn.Flags = %OFN_FILEMUSTEXIST OR %OFN_READONLY OR %OFN_PATHMUSTEXIST
        IF GetOpenFileNameA(ofn) THEN
             KillTimer(gn_hMain, 1)
             LoadModel(szFile)
             ResetCamera()
             CALL SetTimer(gn_hMain, 1, 0, 0)
        END IF

    CASE %MENU_FILE_CLOSE:
'        UnloadModel()

    CASE %MENU_FILE_EXIT:
        SendMessage(hWnd, %WM_CLOSE, 0, 0)

    CASE %MENU_VIEW_FULLSCREEN:
        ToggleFullScreen()
        IF (gn_IsFullScreen) THEN
            CheckMenuItem(GetMenu(hWnd), %MENU_VIEW_FULLSCREEN, %MF_CHECKED)
        ELSE
            CheckMenuItem(GetMenu(hWnd), %MENU_VIEW_FULLSCREEN, %MF_UNCHECKED)
        END IF

    CASE %MENU_VIEW_RESET:
        ResetCamera()

    CASE %MENU_VIEW_CULLBACKFACES:
        gn_CullBackFaces = NOT gn_CullBackFaces
        IF gn_CullBackFaces THEN
            glEnable(%GL_CULL_FACE)
            CheckMenuItem(GetMenu(hWnd), %MENU_VIEW_CULLBACKFACES, %MF_CHECKED)
        ELSE
            glDisable(%GL_CULL_FACE)
            CheckMenuItem(GetMenu(hWnd), %MENU_VIEW_CULLBACKFACES, %MF_UNCHECKED)
        END IF

    CASE %MENU_VIEW_TEXTURED:
        LOCAL nEnableTextures AS LONG
        nEnableTextures = NOT Mobj_enableTextures(0, 0)
        Mobj_enableTextures(nEnableTextures, 1) 
        IF nEnableTextures THEN
            CheckMenuItem(GetMenu(hWnd), %MENU_VIEW_TEXTURED, %MF_CHECKED)
        ELSE
            CheckMenuItem(GetMenu(hWnd), %MENU_VIEW_TEXTURED, %MF_UNCHECKED)
        END IF

    CASE %MENU_VIEW_WIREFRAME:
        gn_EnableWireframe = NOT gn_EnableWireframe
        IF gn_EnableWireframe THEN
            glPolygonMode(%GL_FRONT_AND_BACK, %GL_LINE)
            CheckMenuItem(GetMenu(hWnd), %MENU_VIEW_WIREFRAME, %MF_CHECKED)
        ELSE
            glPolygonMode(%GL_FRONT_AND_BACK, %GL_FILL)
            CheckMenuItem(GetMenu(hWnd), %MENU_VIEW_WIREFRAME, %MF_UNCHECKED)
        END IF
 
    CASE %MENU_SPECULAR_MODE:
        gn_SpecularMode = NOT gn_SpecularMode
        IF gn_SpecularMode THEN
            CheckMenuItem(GetMenu(hWnd), %MENU_SPECULAR_MODE, %MF_CHECKED)
        ELSE
            CheckMenuItem(GetMenu(hWnd), %MENU_SPECULAR_MODE, %MF_UNCHECKED)
        END IF

    CASE %MENU_FIXED_PIPELINE:
        IF gn_SupportsProgrammablePipeline THEN
           IF GetMenuState(GetMenu(hWnd), %MENU_FIXED_PIPELINE, %MF_BYCOMMAND) AND %MF_CHECKED = %MF_CHECKED THEN
              CheckMenuItem(GetMenu(hWnd), %MENU_FIXED_PIPELINE, %MF_UNCHECKED)
           ELSE
              CheckMenuItem(GetMenu(hWnd), %MENU_FIXED_PIPELINE, %MF_CHECKED)
           END IF
        END IF

    CASE %MENU_SWAP_VECTOR:
        gn_SwapVector = NOT gn_SwapVector
        IF gn_SwapVector THEN
            CheckMenuItem(GetMenu(hWnd), %MENU_SWAP_VECTOR, %MF_CHECKED)
        ELSE
            CheckMenuItem(GetMenu(hWnd), %MENU_SWAP_VECTOR, %MF_UNCHECKED)
        END IF
    END SELECT

END SUB




FUNCTION glProc (BYVAL hWnd AS LONG, BYVAL Msg AS LONG, BYVAL wParam AS LONG, BYVAL lParam AS LONG) AS LONG
'    LOCAL ps AS PAINTSTRUCT
'    LOCAL glDC, glRC, lRet AS LONG

    SELECT CASE LONG (msg)

     CASE %WM_SIZE
        gn_WindowWidth = LO(WORD, lParam)
        gn_WindowHeight = HI(WORD, lParam)
        glViewport(0, 0, gn_WindowWidth, gn_WindowHeight)

        '// Part of the code that has been moved from DrawFrame
        glMatrixMode(%GL_PROJECTION)
        glLoadIdentity()
        gluPerspective(CAMERA_FOVY, gn_WindowWidth / gn_WindowHeight, CAMERA_ZNEAR, CAMERA_ZFAR)
        glMatrixMode(%GL_MODELVIEW)

    CASE ELSE
        ProcessMouseInput(hWnd, msg, wParam, lParam)

    END SELECT
    FUNCTION = DefWindowProc (hWnd, Msg, wParam, lParam)

END FUNCTION

FUNCTION Mobj_CreateWindow ALIAS "Mobj_CreateWindow" (BYVAL dwExStyle AS DWORD, _
                                                    BYVAL dwStyle AS DWORD, _
                                                    BYVAL x AS LONG, BYVAL y AS LONG, _
                                                    BYVAL nWidth AS LONG, BYVAL nHeight AS LONG, _
                                                    BYVAL hWndParent AS DWORD, _
                                                    BYVAL CtrlID AS DWORD) EXPORT AS DWORD

    LOCAL IsInitialized, wcStyle AS LONG
    LOCAL hWnd AS DWORD
    LOCAL wc AS WndClassEx
    LOCAL zClass AS ASCIIZ * 16
    zClass = "GLWavefront"

    wcStyle = %CS_HREDRAW OR %CS_VREDRAW OR %CS_DBLCLKS OR %CS_OWNDC
    wc.cbSize = SIZEOF(wc)
    IsInitialized = GetClassInfoEx(zInstance(), zClass, wc)
    IF IsInitialized   = 0 THEN
       wc.cbSize        = SIZEOF(wc)
       wc.style         = wcStyle&
       wc.lpfnWndProc   = CODEPTR(glProc)
       wc.cbClsExtra    = 0
       wc.cbWndExtra    = %Extend_cbWndExtra * 4
       wc.hInstance     = zInstance()
       wc.hIcon         = %NULL
       wc.hCursor       = LoadCursor(%NULL, BYVAL %IDC_ARROW)
       wc.hbrBackground = %NULL
       wc.lpszMenuName  = %NULL
       wc.lpszClassName = VARPTR(zClass)
       wc.hIconSm       = %NULL
       IF RegisterClassEx(wc) THEN IsInitialized = -1
    END IF
    
    IF IsInitialized THEN
    
       hWnd = CreateWindowEx(dwExStyle, _
                             zClass, _         ' Make it an OpenGL control
                             "", _             ' Currently not used
                             dwStyle, _        ' window style
                             x, _              ' initial x position
                             y, _              ' initial y position
                             nWidth, _         ' Calculate Window Width
                             nHeight, _        ' Calculate Window Height
                             hWndParent, _     ' parent window handle
                             CtrlID, _         ' ControlID
                             zInstance(), _    ' program instance handle
                             BYVAL 0)          ' creation parameters
    
    END IF

    FUNCTION = hWnd
END FUNCTION

SUB ProcessMouseInput(BYVAL hWnd AS DWORD, BYVAL msg AS DWORD, BYVAL wParam AS LONG, BYVAL lParam AS LONG)

    '// Use the left mouse button to track the camera.
    '// Use the middle mouse button to dolly the camera.
    '// Use the right mouse button to orbit the camera.

    IF NOT gn_ObjectLoaded THEN EXIT SUB

    STATIC CameraMode, mouseButtonsDown AS LONG
    STATIC ptMousePrev AS POINTAPI
    STATIC ptMouseCurrent AS POINTAPI
    STATIC rx, ry AS SINGLE

    SELECT CASE LONG msg

    CASE %WM_LBUTTONDOWN:
        IF GetFocus() <> hWnd THEN SetFocus(hWnd) 
        cameraMode = %CAMERA_ORBIT
        INCR mouseButtonsDown
        SetCapture(hWnd)
        ptMousePrev.x = LO(INTEGER, lParam)
        ptMousePrev.y = HI(INTEGER, lParam)
        ClientToScreen(hWnd, ptMousePrev)

    CASE %WM_RBUTTONDOWN:
        cameraMode = %CAMERA_TRACK
        INCR mouseButtonsDown
        SetCapture(hWnd)
        ptMousePrev.x = LO(INTEGER, lParam)
        ptMousePrev.y = HI(INTEGER, lParam)
        ClientToScreen(hWnd, ptMousePrev)

    CASE %WM_MBUTTONDOWN:
        cameraMode = %CAMERA_DOLLY
        INCR mouseButtonsDown
        SetCapture(hWnd)
        ptMousePrev.x = LO(INTEGER, lParam)
        ptMousePrev.y = HI(INTEGER, lParam)
        ClientToScreen(hWnd, ptMousePrev)
        IF GetFocus() <> hWnd THEN SetFocus(hWnd)

    CASE %WM_MOUSEWHEEL:
        ry = CSNG(HI(INTEGER, wParam)) / 120 * gr_cameraPos(2) / 20
        gr_cameraPos(2) -= ry
        IF (gr_cameraPos(2) < Mobj_getRadius() + CAMERA_ZNEAR) THEN gr_cameraPos(2) = Mobj_getRadius() + CAMERA_ZNEAR
        IF (gr_cameraPos(2) > CAMERA_ZFAR - Mobj_getRadius()) THEN gr_cameraPos(2) = CAMERA_ZFAR - Mobj_getRadius
        IF GetFocus() <> hWnd THEN SetFocus(hWnd)

    CASE %WM_MOUSEMOVE:
        ptMouseCurrent.x = LO(INTEGER, lParam)
        ptMouseCurrent.y = HI(INTEGER, lParam)
        ClientToScreen(hWnd, ptMouseCurrent)

        SELECT CASE LONG cameraMode
        CASE %CAMERA_TRACK:
            rx = (ptMouseCurrent.x - ptMousePrev.x) * MOUSE_TRACK_SPEED
            ry = (ptMouseCurrent.y - ptMousePrev.y) * MOUSE_TRACK_SPEED

            gr_cameraPos(0) -= rx
            gr_cameraPos(1) += ry

            gr_targetPos(0) -= rx
            gr_targetPos(1) += ry

        CASE %CAMERA_DOLLY:
            ry = (ptMouseCurrent.y - ptMousePrev.y) * MOUSE_DOLLY_SPEED

            gr_cameraPos(2) -= ry

            IF (gr_cameraPos(2) < Mobj_getRadius() + CAMERA_ZNEAR) THEN gr_cameraPos(2) = Mobj_getRadius() + CAMERA_ZNEAR
            IF (gr_cameraPos(2) > CAMERA_ZFAR - Mobj_getRadius()) THEN gr_cameraPos(2) = CAMERA_ZFAR - Mobj_getRadius()

        CASE %CAMERA_ORBIT:
            rx = (ptMouseCurrent.x - ptMousePrev.x) * MOUSE_ORBIT_SPEED
            ry = (ptMouseCurrent.y - ptMousePrev.y) * MOUSE_ORBIT_SPEED

            gr_heading += rx
            gr_pitch += ry

'zizi            IF gr_pitch > 90.0 THEN gr_pitch = 90.0
'zizi            IF gr_pitch < -90.0 THEN gr_pitch = -90.0

        END SELECT

        ptMousePrev.x = ptMouseCurrent.x
        ptMousePrev.y = ptMouseCurrent.y

    CASE %WM_LBUTTONUP, %WM_RBUTTONUP, %WM_MBUTTONUP:
        DECR mouseButtonsDown
        IF (mouseButtonsDown < 1) THEN
            mouseButtonsDown = 0
            cameraMode = %CAMERA_NONE
            ReleaseCapture()
        ELSE
            IF (wParam AND %MK_LBUTTON) THEN
               cameraMode = %CAMERA_TRACK
            ELSEIF (wParam AND %MK_RBUTTON) THEN
               cameraMode = %CAMERA_ORBIT
            ELSEIF (wParam AND %MK_MBUTTON) THEN
               cameraMode = %CAMERA_DOLLY
            END IF
        END IF

    CASE ELSE

    END SELECT

END SUB
